
from __future__ import annotations
import typing 

if typing.TYPE_CHECKING:
    from discordoauth.scopes import Scopes

from discordoauth import session as s
from discordoauth import http

import asyncio
import nest_asyncio

class Object():
    def __init__(self, id : int):
        self.id = id 
    
    def __int__(self):
        return self.id

class Client():
    def __init__(self, app_id : int, app_secret : str):
        self.app_id = app_id 
        self.app_secret = app_secret

        self.loop = asyncio.new_event_loop()
        nest_asyncio.apply(self.loop)

        self.http = http.HTTP(self)
    
    def new_session(self, code : str, scopes : Scopes, rediret_uri=None) -> s.Session:

        session = s.Session(self, code, scopes, rediret_uri)

        return session
    
    def get_oauth_url(self, scopes : Scopes, rediret_uri=None):

        url = f"{self.http.base}/oauth2/authorize?client_id={self.app_id}"

        if rediret_uri:
            url += f"&rediret_uri={rediret_uri}"
        
        url += f"&scope={str(scopes)}"

        if scopes.identify:
            url += "&response_type=code"

        return url



from __future__ import annotations
import typing 

if typing.TYPE_CHECKING:
    from discordoauth.client import Client, Object as ObjectType
    from discordoauth.scopes import Scopes
    from discordoauth.session import Session as SessionType, Member as MemberType

from discordoauth import scopes, errors
from discordoauth import utils

class Access():
    def __init__(self, data : dict):
        self.access_token : str = data["access_token"]
        self.expires_in : int = data["expires_in"]
        self.refresh_token : str= data["refresh_token"]

        self.scopes = scopes.Scopes().from_string(data["scope"])

class User():

    def __init__(self, data : dict):

        self.raw = data

        self.id = int(data["id"])
        self.name : str = data["username"]
        self.avatar_url = f"https://cdn.discordapp.com/avatars/{self.id}/{data['avatar']}.webp?size=1024"
        self.discriminator = int(data["discriminator"])
        self.banner : str = data["banner"] if "banner" in data else None
        self.banner_color : str = data["banner_color"] if "banner_color" in data else None
        self.locale : str = data["locale"] if "locale" in data else None
        self.mfa_enabled : bool = data["mfa_enabled"] if "mfa_enabled" in data else None

        if "email" in data:
            self.verified : bool = data["verified"]
            self.email : str = data["email"]
    
    def __str__(self):
        return f"{self.name}#{self.discriminator}"
    
    def __dict__(self):
        return self.raw

class Guild():
    def __init__(self, session : SessionType, data : dict):

        self.raw = data
        self.session = session
        
        self.id = int(data["id"])
        self.name : str = data["name"]
        self.icon : str = f"https://cdn.discordapp.com/icons/{self.id}/{data['icon']}.webp?size=1024"
        self.is_owner : bool = data["owner"]
        self.features : typing.List[str] = data["features"]

        self.member : MemberType = None
    
    async def fetch_member_information(self):
        
        if not self.session.scopes.guilds_members_read:
            raise errors.MissingScope("This method requires the guilds_members_read scope")

        if not self.session.access:
            await self.session.get_access_token()
        
        headers = {"Content-Type":"application/json", 'Authorization': "Bearer {}".format(self.session.access.access_token)}

        data = await self.session.client.http.request_rate(self.session, "GET", f"/users/@me/guilds/{self.id}/member", headers=headers)

        self.member = Member(data) 

        return self.member
    
    def __str__(self):
        return self.name 
    
    def __dict__(self):
        return self.raw

class Connection():
    def __init__(self, data : dict):

        self.raw = data
        
        self.type : str = data["type"]
        self.id : str = data["id"]
        self.name : str = data["name"]
        self.is_visible = bool(data["visibility"])
        self.friend_sync : bool = data["friend_sync"]
        self.show_activity : bool = data["show_activity"]
        self.id_verified : bool = data["verified"]
    
    def __str__(self):
        return self.name 
    
    def __dict__(self):
        return self.raw

class Member():
    def __init__(self, data : dict):

        self.raw = data
        
        self.user = User(data["user"])
        self.avatar_url : str = f"https://cdn.discordapp.com/avatars/{self.user.id}/{data['avatar']}.webp?size=1024"

        self.communication_disabled_until : str = data["communication_disabled_until"]
        self.flags : int = data["flags"]
        self.is_pending : bool = data["is_pending"]
        self.joined_at : str = data["joined_at"]
        self.nick : str = data["nick"]
        self.premium_since : str = data["premium_since"]
        self.roles : typing.List[dict] = data["roles"]
        self.is_mute : bool = data["mute"]
        self.is_deaf : bool = data["deaf"]
    
    def __str__(self):
        return f"{self.user.name}#{self.user.discriminator}"
    
    def __dict__(self):
        return self.raw

class Session():

    def __init__(self, client : Client, code : str, scopes : Scopes, rediret_uri : str = None):

        self.client = client 

        self.code = code 
        self.scopes = scopes 
        self.redirect_uri = rediret_uri

        self.access : Access = None

        self.user : User = None 
        self.guilds : typing.List[Guild] = []
        self.connections : typing.List[Connection] = []
        self.member : Member = None
    
    def get_access_token_sync(self) -> Access:
        return self.client.loop.run_until_complete(self.get_access_token())
    
    def refresh_access_token_sync(self, refresh_token : str = None) -> Access:
        return self.client.loop.run_until_complete(self.refresh_access_token(refresh_token))
    
    def fetch_user_sync(self) -> User:
        return self.client.loop.run_until_complete(self.fetch_user())
    
    def fetch_guilds_sync(self) -> typing.List[Guild]:
        return self.client.loop.run_until_complete(self.fetch_guilds())
    
    def fetch_connections_sync(self) -> typing.List[Connection]:
        return self.client.loop.run_until_complete(self.fetch_connections())
    
    def join_guild_sync(
            self, 
            bot_token : str, 
            guild_id : int, 
            nick : str = None, 
            roles : typing.List[typing.Union[ObjectType, int]] = None, 
            mute : bool = False,
            deaf : bool = False,
            user_id : int = None
    ) -> typing.Union[Member, None]:
        return self.client.loop.run_until_complete(self.join_guild(bot_token, guild_id, nick, roles, mute, deaf, user_id))
    
    
    async def get_access_token(self) -> Access:

        """
            Retrives an oauth access token from discord

            This must be called before retreiving any data.
        """

        payload = {
            'client_id': self.client.app_id,
            'client_secret': self.client.app_secret,
            'grant_type': 'authorization_code',
            'code': self.code,
            'redirect_uri': self.redirect_uri,
            'scope': str(self.scopes)
        }

        data = await self.client.http.request_rate(self, "POST", "/oauth2/token", payload, {'Content-Type': 'application/x-www-form-urlencoded'})
        self.access = Access(data)

        return self.access
    
    async def refresh_access_token(self, refresh_token : str = None) -> Access:

        """
            Retrives an oauth access token from discord

            This must be called before retreiving any data.
        """

        refresh_token = refresh_token or self.access.refresh_token

        payload = {
            'client_id': self.client.app_id,
            'client_secret': self.client.app_secret,
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }

        data = await self.client.http.request_rate(self, "POST", "/oauth2/token", payload, {'Content-Type': 'application/x-www-form-urlencoded'})
        self.access = Access(data)

        return self.access
    
    async def fetch_user(self) -> User:

        if not self.scopes.identify:
            raise errors.MissingScope("This method requires the identify scope")

        if not self.access:
            await self.get_access_token()
        
        headers = {
            'Authorization': "Bearer {}".format(self.access.access_token)
        }

        data = await self.client.http.request_rate(self, "GET", "/users/@me", headers=headers)
        self.user = User(data)

        return self.user
    
    async def fetch_guilds(self) -> typing.List[Guild]:

        if not self.scopes.guilds:
            raise errors.MissingScope("This method requires the guilds scope")

        if not self.access:
            await self.get_access_token()
        
        headers = {
            'Authorization': "Bearer {}".format(self.access.access_token)
        }

        data = await self.client.http.request_rate(self, "GET", "/users/@me/guilds", headers=headers)
        self.guilds : typing.List[Guild] = []

        for raw_guild in data:     
            self.guilds.append(Guild(self, raw_guild))   
        
        return self.guilds
    
    async def fetch_connections(self) -> typing.List[Connection]:

        if not self.scopes.connections:
            raise errors.MissingScope("This method requires the connections scope")

        if not self.access:
            await self.get_access_token()
        
        headers = {
            'Authorization': "Bearer {}".format(self.access.access_token)
        }

        data = await self.client.http.request_rate(self, "GET", "/users/@me/connections", headers=headers)
        self.connections : typing.List[Connection] = []

        for raw_connection in data:     
            self.connections.append(Connection(raw_connection))   
        
        return self.connections
    
    async def join_guild(
            self, 
            bot_token : str, 
            guild_id : int, 
            nick : str = None, 
            roles : typing.List[typing.Union[ObjectType, int]] = None, 
            mute : bool = False,
            deaf : bool = False,
            user_id : int = None
    ) -> typing.Union[Member, None]:

        if not self.scopes.guilds_join:
            raise errors.MissingScope("This method requires the guilds_join scope")
        
        if not user_id and not self.user:
            raise errors.MissingArgument("Argument user_id is missing and no previous user data was found.")

        if hasattr(roles, '__iter__'):
            for i, role in enumerate(roles):
                if hasattr(role, "id"):
                    roles[i] = str(role.id)
                else:
                    roles[i] = str(role)
        
        user_id = user_id or self.user.id

        if not self.access:
            await self.get_access_token()
        
        headers = {
            'Authorization': "Bot {}".format(bot_token),
            "Content-Type":"application/json"
        }

        data = {
            "access_token":self.access.access_token,
            "nick":nick,
            "deaf":deaf,
            "mute":mute,
            "roles":roles
        }



        data = await self.client.http.request_rate(self, "PUT", f"/guilds/{guild_id}/members/{user_id}", json=data, headers=headers)

        if data:
            self.member = Member(data) 

            return self.member
    
    def get_guild(self, guild_id : int) -> typing.Union[Guild, None]:

        return utils.item_by_value(self.guilds, "id", guild_id)
    

    


from __future__ import annotations
import typing 

if typing.TYPE_CHECKING:
    from discordoauth.client import Client
    from discordoauth.session import Session

import aiohttp 
import json as j

import asyncio
from discordoauth import errors

INFINITY = 9999999999

class HTTP():

    def __init__(self, client : Client):

        self.client = client 
        
        self.base = "https://discord.com/api"

        self.session : aiohttp.ClientSession = None
    
    async def get_aiohttp_session(self) -> aiohttp.ClientSession:

        if not self.session:
            self.session = aiohttp.ClientSession()

        return self.session 
    
    async def request_rate(self, oauth_session : Session, request : str, path : str, data : dict = None, headers : dict =None, json : dict = None):

        session = await self.get_aiohttp_session()
        retry_after = INFINITY
        tried_refresh = False

        # Rate limit handling.
        while retry_after > 0:
            async with session.request(request, self.base + path, data=data, headers=headers, json=json) as r:

                if r.status == 400 and not tried_refresh:
                    await oauth_session.refresh_access_token()
                    tried_refresh = True
                if r.status == 429:
                    retry_after = ((await r.json())["retry_after"] / 1000) + 0.5
                    await asyncio.sleep(retry_after + 1)
                
                retry_after = 0  

                if r.status == 204:
                    return None
                
                data = j.loads(await r.text())

                if r.status == 403:
                    raise errors.HTTPError(data["message"])

                return data


        





from setuptools import setup, Extension
import discordoauth

version = discordoauth.__version__

setup(
  name = 'discordoauth.py',         
  packages = ['discordoauth'],   
  project_urls={
        "Documentation": "https://github.com/Coolo22/discordoauth.py/docs",
        "Issue tracker": "https://github.com/Coolo22/discordoauth.py/issues",
},
  version = version,     
  license='MIT',       
  description = 'Easily using discord webhooks in python - asynchronous and synchronous - documented at https://github.com/Coolo22/discordoauth.py/docs/', 
  long_description=open("README.md").read(),
  long_description_content_type='text/markdown',
  author = 'Coolo2',                   
  author_email = 'itsxcoolo2@gmail.com',      
  url = 'https://github.com/Coolo22/discordoauth.py',   
  download_url = 'https://github.com/Coolo22/discordwebhook.py/raw/master/archive/discordoauth.py-' + version + '.tar.gz',    
  keywords = ['discord', 'oauth2', 'python', 'api', 'asynchronous', 'synchronous', "oauth"],   
  install_requires=['aiohttp', 'nest_asyncio'],
  classifiers=[
    'Development Status :: 5 - Production/Stable', 
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   
    'Programming Language :: Python :: 3',      
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
  ],
)




VALID_SCOPES = ["bot", "identify", "guilds", "connections", "email", "guilds_join", "guilds_members_read", "applications_commands"]

class Scopes():

    """
        Scopes - A list of Discord Oauth2 scopes
        
        Supported scopes: bot, identify, guilds, connections, email, guilds_join, guilds_members_read, applications_commands

        Example:
        `discordoauth.Scopes(bot=True, identify=True)`
    """    

    def __init__(self, **scopes):

        self.bot = False 
        self.identify = False
        self.guilds = False
        self.connections = False
        self.email = False
        self.guilds_join = False
        self.guilds_members_read = False
        self.applications_commands = False

        self.scopes = []
        
        for key, value in scopes.items():
            if key not in VALID_SCOPES:
                raise TypeError(f'{key!r} is not a valid scope name')
            
            if value:
                self.scopes.append(key.replace("_", "."))
            
            setattr(self, key, value)
    
    def __str__(self):
        return "%20".join(self.scopes)
    
    
    def from_string(self, string : str):
        
        scope_list = string.split("%20")

        for scope in scope_list:
            setattr(self, scope, True)
        self.scopes = scope_list



import typing

def append_if_doesnt_exist(array : list, appendable : typing.Any, parameter : str) -> list:

    exists = False
    for item in array:
        if getattr(appendable, parameter) == getattr(item, parameter):
            exists = True 
    
    if not exists:
        array.append(appendable)
    
    return array

def item_by_value(array : list, parameter : str, value : typing.Any):
    exists = None
    for item in array:
        if getattr(item, parameter) == value:
            exists = item 
    
    return exists